package com.unitesk.requality.libreoffice.api;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;

import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.databind.BeanProperty;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.JavaType;
import com.fasterxml.jackson.databind.JsonDeserializer;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.deser.ContextualDeserializer;
import com.fasterxml.jackson.databind.module.SimpleDeserializers;
import com.fasterxml.jackson.databind.module.SimpleModule;
import com.fasterxml.jackson.databind.node.TextNode;
import com.sun.star.lang.IllegalArgumentException;
import com.unitesk.requality.libreoffice.api.requests.PingRequest;
import com.unitesk.requality.libreoffice.api.responses.PingResponse;

public class APIClient {
	public static final String HARDCODED_URL = "http://localhost:9988/";			// TODO
	public static final String HARDCODED_KEY = "36AC0A0F7F271314B72986EB54DB2343";	// TODO
	
	private ObjectMapper jsonMapper = new ObjectMapper();
	private URL baseUrl;
	private String apiKey;
	private boolean debug;
	
	private class WrappedResponseReader extends JsonDeserializer<WrappedResponse<?>> implements ContextualDeserializer {
		private JavaType underlyingType;
		
		@Override
		public WrappedResponse<?> deserialize(JsonParser parser, DeserializationContext ctx) 
				throws IOException, JsonProcessingException {
			if (parser.nextToken() != JsonToken.FIELD_NAME) {
				throw new IllegalArgumentException("Field name expected, got " + parser.currentToken());
			}
			
			if (!parser.getText().equals("success")) {
				throw new IllegalArgumentException("Success field expected");
			}
			
			parser.nextToken();
			boolean success = parser.getBooleanValue();
			
			if (parser.nextToken() != JsonToken.FIELD_NAME) {
				throw new IllegalArgumentException("Field name expected, got " + parser.currentToken());
			}
			
			if (!parser.getText().equals(success ? "result" : "error")) {
				throw new IllegalArgumentException("Data field expected");
			}
			
			JsonToken openToken = null, closeToken = null;
			if (!success || underlyingType.isTypeOrSubTypeOf(APIResponse.class)) {
				openToken = JsonToken.START_OBJECT;
				closeToken = JsonToken.END_OBJECT;
			} else if (underlyingType.isTypeOrSubTypeOf(TextNode.class)) {
				openToken = JsonToken.VALUE_STRING;
			}
			
			if (openToken != null && parser.nextToken() != openToken) {
				throw new IllegalArgumentException(openToken + " expected, got " + parser.currentToken());
			}
			
			WrappedResponse<?> resp;
			if (success) {
				resp = new WrappedResponse<Object>(ctx.readValue(parser, underlyingType));
			} else {
				resp = new WrappedResponse<Object>(ctx.readValue(parser, RequestError.class));
			}
			
			if (closeToken != null && parser.nextToken() != closeToken) {
				throw new IllegalArgumentException(closeToken + " expected, got " + parser.currentToken());
			}
			
			return resp;
		}

		@Override
		public JsonDeserializer<?> createContextual(DeserializationContext arg0, BeanProperty arg1) 
				throws JsonMappingException {
			WrappedResponseReader reader = new WrappedResponseReader();
			reader.underlyingType = (arg1 == null ? arg0.getContextualType() : arg1.getType()).containedType(0);
			return reader;
		}
	}
	
	private static boolean shouldDebug() {
		String debugEnv = System.getenv("LOREQUALITY_API_DEBUG");
		return debugEnv != null && debugEnv.toLowerCase().equals("true");
	}
	
	public APIClient(URL baseUrl, String apiKey) throws IOException {
		this(baseUrl, apiKey, shouldDebug());
	}
	
	public APIClient(URL baseUrl, String apiKey, boolean debug) throws IOException {
		this.baseUrl = baseUrl;
		this.apiKey = apiKey;
		this.debug = debug;
		
		jsonMapper.enable(SerializationFeature.INDENT_OUTPUT);
		jsonMapper.disable(SerializationFeature.FAIL_ON_EMPTY_BEANS);
		jsonMapper.registerModule(new SimpleModule() {
			private static final long serialVersionUID = 1L;

			@Override
			public void setupModule(SetupContext ctx) {
				super.setupModule(ctx);
				
				Map<Class<?>, JsonDeserializer<?>> deserializers = new HashMap<>();
				
				deserializers.put(WrappedResponse.class, new WrappedResponseReader());
				
				ctx.addDeserializers(new SimpleDeserializers(deserializers));
			}
		});
	}
	
	public URL getBaseURL() {
		return baseUrl;
	}

	public String getApiKey() {
		return apiKey;
	}

	public boolean isDebug() {
		return debug;
	}

	public void setDebug(boolean debug) {
		this.debug = debug;
	}

	public void testPing() throws IOException {
		PingRequest ping = PingRequest.random();
		PingResponse pingRes = execute(ping);
		if (!ping.matches(pingRes))
			throw new IOException("API fails ping testing: " + ping + ", " + pingRes);
	}

	public <R, T extends APIRequest<R>> R execute(T request) throws IOException {
		long start = System.currentTimeMillis();
		
		HttpURLConnection connection = (HttpURLConnection) new URL(baseUrl, request.requestURL()).openConnection();
		connection.setRequestMethod(request.method().name());
		connection.setRequestProperty("Authorization", "Bearer " + apiKey);
		
		if (request.method().hasBody()) {
			connection.setDoOutput(true);
			connection.setRequestProperty("Content-Type", "application/json");
			
			OutputStream os = connection.getOutputStream();
			if (debug) {
				String content = jsonMapper.writeValueAsString(request);
				System.out.println("<<<< START");
				System.out.println(content);
				System.out.println("<<<< END");
				os.write(content.getBytes());
			} else {
				jsonMapper.writeValue(os, request);
			}
			os.close();
		}
		
		JavaType resultType = 
				jsonMapper.getTypeFactory().constructParametricType(WrappedResponse.class, request.responseClass());
		
		WrappedResponse<R> resp;
		InputStream input;
		
		try {
			input = connection.getInputStream();
		} catch (IOException e) {
			input = connection.getErrorStream();
			
			if (input == null) {
				throw e;
			}
		}
		
		try {
			if (debug) {
				StringWriter sw = new StringWriter();
				InputStreamReader isr = new InputStreamReader(input, StandardCharsets.UTF_8);
				char[] buffer = new char[128];
				int r;
				
				while ((r = isr.read(buffer)) != -1) {
					sw.write(buffer, 0, r);
				}
				
				String json = sw.toString();
				
				System.out.println(">>>> START");
				System.out.println(json);
				System.out.println(">>>> END");
				
				resp = jsonMapper.readValue(json, resultType);
			} else {
				resp = jsonMapper.readValue(input, resultType);
			}
		} finally {
			input.close();
		}
		
		int time = (int) (System.currentTimeMillis() - start);
		
		if (debug) {
			System.out.println(">>>> PARSED START");
			System.out.println(resp);
			System.out.println(">>>> PARSED END (" + time + " ms)");
		} else {
			System.out.println(request.getClass().getCanonicalName() + ": " + time + " ms");
		}
		
		if (resp.hasError()) {
			throw new RequestException(resp.getError());
		} else {
			return resp.getResponse();
		}
	}
	
	public static APIClient get() throws IOException {
		return new APIClient(new URL(HARDCODED_URL), HARDCODED_KEY);
	}
}
